from .base import Base
import logging
from pika.exceptions import ConnectionClosed
LOGGER = logging.getLogger(__name__)


class Emitter(Base):
    def emit(self, key, message):
        """
         Emits a message to RabbitMQ

        :param key: Routing key to emit on
        :param message: json serialized msg body
        :return:
        :raises Exception: If RAISE_EXCEPTION is true in emitter settings
        """
        try:
            self._emit(key, message)
        except ConnectionClosed:
            self.shutdown()
            self._emit(key, message)
        except Exception as e:
            LOGGER.exception("Emitter exception %s" % str(e))
            if self.settings.EMITTER['RAISE_EXCEPTION']:
                raise

    def _emit(self, key, message):
        pub_channel = self.get_channel()
        LOGGER.debug("Trying to emit - %s" % message)
        if pub_channel.basic_publish(exchange=self.settings.EXCHANGE, routing_key=key, body=message):
            LOGGER.info("Emitted - %s" % message)
        else:
            LOGGER.error("Failed to emit - %s" % message)



import json

from django.db.models.signals import post_save, post_delete
from django.core import serializers
from .settings import project_name, MODEL_DEFAULTS
from hedwig.core.settings import Settings
from .emitter import hedwig_emitter
import inspect

import logging

LOGGER = logging.getLogger(__name__)


def coerce_model_settings(instance):
    model_hedwig_dict = {}
    model_hedwig = getattr(instance, 'Hedwig', None)
    if inspect.isclass(model_hedwig):
        model_hedwig_dict = dict((name, getattr(model_hedwig, name)) for name in dir(model_hedwig) if not name.startswith('__'))
    return Settings(user_settings=model_hedwig_dict, defaults=MODEL_DEFAULTS)


def saved_hedwig_models(sender, instance, created, **kwargs):
    if created:
        action = 'created'
    else:
        action = 'updated'
    emit_hedwig_model_data(instance, action)


def deleted_hedwig_models(sender, instance, **kwargs):
    emit_hedwig_model_data(instance, 'deleted')


def emit_hedwig_model_data(instance, action):
    model_settings = coerce_model_settings(instance)
    if not getattr(model_settings, action, True):
        return True

    routing_key = '.'.join([project_name, instance._meta.app_label, 'model', instance._meta.object_name, action,
                            str(instance.pk)])
    show_fields = True
    serializer_kwargs = {}
    if isinstance(model_settings.fields, bool):
        show_fields = model_settings.fields
    else:
        serializer_kwargs['fields'] = model_settings.fields
    hedwig_json_obj = {}
    if show_fields:
        serializer_kwargs['use_natural_primary_keys'] = model_settings.natural_primary_keys
        serializer_kwargs['use_natural_foreign_keys'] = model_settings.natural_foreign_keys
        serializer_kwargs['format'] = "json"
        serializer_kwargs['queryset'] = [instance]
        try:
            hedwig_json_obj = json.loads(serializers.serialize(**serializer_kwargs))[0]
        except Exception as e:

            hedwig_json_obj = {'error': str(e)}

    payload = json.dumps(hedwig_json_obj)

    hedwig_emitter.emit(routing_key, payload)


def register_hedwig_callbacks():
    LOGGER.info("Registering post_save and post_delete signal callbacks")
    post_save.connect(saved_hedwig_models, weak=False, dispatch_uid="saved_hedwig_models")
    post_delete.connect(deleted_hedwig_models, weak=False, dispatch_uid="deleted_hedwig_models")



from future.utils import iteritems
from copy import deepcopy

DEFAULTS = {
    'EXCHANGE': 'hedwig',
    'EXCHANGE_TYPE': 'topic',
    'HEARTBEAT_INTERVAL': 5,
    'SOCKET_TIMEOUT': 1,
    'HOST': 'localhost',
    'PORT': 5672,
    'VHOST': '',
    'USERNAME': '',
    'PASSWORD': '',
    'EMITTER': {
        'RAISE_EXCEPTION': False,
    },
    'CONSUMER': {
        'RAISE_EXCEPTION': False,
        'QUEUES': {}
    }
}

# Nod to https://www.xormedia.com/recursively-merge-dictionaries-in-python/


def dict_merge(a, b):
    """recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and bhave a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary."""
    if not isinstance(b, dict):
        return b
    result = deepcopy(a)
    for k, v in iteritems(b):
        if k in result and isinstance(result[k], dict):
                result[k] = dict_merge(result[k], v)
        else:
            result[k] = deepcopy(v)
    return result


# Nod to DRF for their settings, added dict_merge with deep copy to make it work with our use case

class Settings(object):
    """
    A settings object, that allows HEDWIG settings to be accessed as properties.
    """

    def __init__(self, user_settings=None, defaults=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or DEFAULTS
        if isinstance(self.user_settings, dict):
            self.user_settings = dict_merge(self.defaults, self.user_settings)
        else:
            raise TypeError("Expected settings to be a dict")

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid HEDWIG setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)
        return val

    def __getitem__(self, item):
        return getattr(self, item)



"""
 (@,@)
                _          _
  /\  /\___  __| |_      _(_) __ _
 / /_/ / _ \/ _` \ \ /\ / / |/ _` |
/ __  /  __/ (_| |\ V  V /| | (_| |
\/ /_/ \___|\__,_| \_/\_/ |_|\__, |
                             |___/
"""

__title__ = 'Hedwig Python'
__version__ = '0.1.4'
__author__ = 'Piyush'

# Version synonym
VERSION = __version__

import logging

try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())



import pika
import logging
from pika.exceptions import ConnectionClosed

LOGGER = logging.getLogger(__name__)


class Base(object):
    def __init__(self, settings):
        LOGGER.info("Base init for class - {0}".format(self.__class__.__name__))
        self.settings = settings
        LOGGER.debug("Setting credentials")
        credentials = pika.PlainCredentials(username=self.settings.USERNAME,
                                            password=self.settings.PASSWORD)
        LOGGER.debug("Setting Connection params")
        self.conn_params = pika.ConnectionParameters(host=self.settings.HOST, port=self.settings.PORT,
                                                     virtual_host=self.settings.VHOST, credentials=credentials,
                                                     heartbeat_interval=self.settings.HEARTBEAT_INTERVAL,
                                                     socket_timeout=self.settings.SOCKET_TIMEOUT)

        self.connection = None
        self.channel = None

    def create_channel(self):
        LOGGER.debug("Creating a channel")
        channel = self.connection.channel()
        LOGGER.debug(
            "Declaring {0} exchange - {1}".format(self.settings.EXCHANGE_TYPE, self.settings.EXCHANGE)
        )
        channel.exchange_declare(exchange=self.settings.EXCHANGE, exchange_type=self.settings.EXCHANGE_TYPE, durable=True)
        return channel

    def get_channel(self):

        self.connect()
        if self.channel is None or not self.channel.is_open:
            LOGGER.info("No active channel found, creating a channel")
            self.channel = self.create_channel()
        return self.channel

    def connect(self):
        LOGGER.debug("Making a blocking connection")
        if self.connection is None or not self.connection.is_open:
            LOGGER.info("No active connection found, creating a connection")
            self.connection = pika.BlockingConnection(self.conn_params)

    def close_channel(self):
        LOGGER.debug("Trying to close channel if open")
        if self.channel and self.channel.is_open:
            LOGGER.debug("Closing channel")
            try:
                self.channel.cancel()
                self.channel.close()
            except ConnectionClosed:
                self.channel = None

    def shutdown(self):
        LOGGER.info("Shutdown called")
        self.close_channel()
        if self.connection and self.connection.is_open:
            LOGGER.info("Closing connection")
            try:
                self.connection.close()
            except ConnectionClosed:
                self.connection = None



import logging

from importlib import import_module

LOGGER = logging.getLogger(__name__)


def import_obj(cb):
    LOGGER.debug("Attempting to import - %s" % cb)
    try:
        # Nod to DRF for nod to tastypie's use of importlib.
        parts = cb.split('.')
        module_path, callback_name = '.'.join(parts[:-1]), parts[-1]
        LOGGER.debug("Importing %s from %s" % (callback_name, module_path))
        module = import_module(module_path)
        return getattr(module, callback_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '%s' %s: %s." % (cb, e.__class__.__name__, e)
        raise ImportError(msg)



from hedwig.core.settings import Settings
from django.conf import settings

REST_FRAMEWORK_DEFAULTS = {
    'SERIALIZER_SIGNALS': True
}

HEDWIG_REST_FRAMEWORK_USER_SETTINGS = getattr(settings, 'HEDWIG_REST_FRAMEWORK', None)

hedwig_rest_framework_settings = Settings(HEDWIG_REST_FRAMEWORK_USER_SETTINGS, REST_FRAMEWORK_DEFAULTS)



from setuptools import setup
import os
import re

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    # Nod to DRF for this function
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


setup(name="hedwig-py",
      version=get_version('hedwig'),
      description="Python emitter and consumer for Hedwig",
      url="https://github.com/ofpiyush/hedwig-py/",
      author="Piyush",
      author_email="mail@ofpiyush.in",
      license='MIT',
      packages=get_packages('hedwig'),
      install_requires=requirements,
      zip_safe=False,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Topic :: System :: Distributed Computing',
      ]
      )



import os

from django.conf import settings
from hedwig.core.settings import Settings, DEFAULTS

HEDWIG_USER_SETTINGS = getattr(settings, 'HEDWIG', None)
HEDWIG_DJANGO_USER_SETTINGS = getattr(settings, 'HEDWIG_DJANGO', None)


DJANGO_DEFAULTS = {
    'MODEL_SIGNALS': False,
    'HIDE_FIELDS': False,
    'NATURAL_PRIMARY_KEYS': True,
    'NATURAL_FOREIGN_KEYS': True
}

hedwig_settings = Settings(HEDWIG_USER_SETTINGS, DEFAULTS)
hedwig_django_settings = Settings(HEDWIG_DJANGO_USER_SETTINGS, DJANGO_DEFAULTS)

project_name = os.getenv('DJANGO_SETTINGS_MODULE').split('.')[0]


MODEL_DEFAULTS = {
    'fields': not bool(hedwig_django_settings.HIDE_FIELDS),
    'natural_primary_keys': bool(hedwig_django_settings.NATURAL_PRIMARY_KEYS),
    'natural_foreign_keys': bool(hedwig_django_settings.NATURAL_FOREIGN_KEYS),
    'created': True,
    'updated': True,
    'deleted': True
}



import logging
from future.utils import iteritems

from pika.exceptions import ConnectionClosed

from hedwig import utils
from hedwig.core.base import Base

LOGGER = logging.getLogger(__name__)


class Consumer(Base):
    def __init__(self, settings):
        self._callbacks = {}
        super(Consumer, self).__init__(settings)

    def consume(self):
        """
        Consumes messages from RabbitMQ
        :return:
        :raises KeyboardInterrupt: if one was sent to it
        :raises Exception: if RAISE_EXCEPTION on CONSUMER is true
        """
        LOGGER.debug('Consumer Initialized')
        # self.connect()
        channel = self.get_channel()
        self._bind_things(channel)

        try:
            LOGGER.info('Start consuming')
            channel.start_consuming()
        except ConnectionClosed:
            LOGGER.exception('Pika connection closed detected. Will attempt to start consuming again')
            self.consume()
        except KeyboardInterrupt as e:
            LOGGER.info('Keyboard interrupt, stop consuming')
            self.shutdown()
            raise
        except Exception as e:
            LOGGER.exception("'%s" % str(e))
            self.shutdown()
            if self.settings.CONSUMER['RAISE_EXCEPTION']:
                LOGGER.info("CONSUMER RAISED EXCEPTION")
                raise

    def callback(self, func_string):
        """
        Wraps callback from settings

        :param func_string: string notation of actual callback to be called
        :return callback_wrapper: Callback wrapper
        :raises Exception: if consumer RAISE_EXCEPTION is true
        """
        def callback_wrapper(ch, method, properties, body):
            LOGGER.info("Got message - %s with body %s" % (method.routing_key, body))
            if func_string not in self._callbacks:
                self._callbacks[func_string] = utils.import_obj(func_string)
            try:
                self._callbacks[func_string](ch, method, properties, body)
            except Exception as e:
                LOGGER.exception("Callback exception '%s'" % str(e))
                if self.settings.CONSUMER['RAISE_EXCEPTION']:
                    LOGGER.info("CALLBACK RAISED EXCEPTION")
                    raise

        return callback_wrapper

    def _bind_things(self, channel):
        LOGGER.debug('Attempting to bind queues')
        for q_name, q_settings in iteritems(self.settings.CONSUMER['QUEUES']):
            LOGGER.debug('Declaring queue - %s' % q_name)
            channel.queue_declare(queue=q_name, durable=q_settings['DURABLE'], auto_delete=q_settings['AUTO_DELETE'])
            for binding in q_settings['BINDINGS']:
                LOGGER.info("Binding the queue - %s with key %s" % (q_name, binding))
                channel.queue_bind(exchange=self.settings.EXCHANGE, queue=q_name,
                                   routing_key=binding)
            LOGGER.debug("Setting Basic consume for callback - %s " % q_settings['CALLBACK'])
            channel.basic_consume(self.callback(q_settings['CALLBACK']), queue=q_name,
                                  no_ack=q_settings['NO_ACK'])




from django.apps.config import AppConfig
from .signals import register_hedwig_callbacks
from .settings import hedwig_django_settings


class HedwigAppConfig(AppConfig):
    name = 'hedwig.django'
    label = 'hedwig.django'

    def ready(self):
        if hedwig_django_settings.MODEL_SIGNALS:
            register_hedwig_callbacks()



__author__ = 'sandeep'
import sys

from hedwig.core.service import ServiceManager
from .worker import DjangoHedwigWorker


def main(argv):
    msg_service = ServiceManager(DjangoHedwigWorker, num_workers=2)
    msg_service.start()


if __name__ == "__main__":
    main(sys.argv[1:])
